from Samples import Samples
import random
import copy
import math
from distance import distance

##MOs are either used as routing objects or dataobjects
class MO:

##Constructor. Sets some variables.
    def __init__(self, position, range, nType):
        self.position = position
        self.range = range
        self.nType = nType
        self.parent = None
        self.node = None
        self.pd = 0

##Sets the key of the object
    def setPosition(self, position):
        self.position = position

##Returns the key of the object
    def getPosition(self):
        return self.position

##Sets the node the object belongs to.
    def setNode(self, node):
        self.node = node

##Returns the node the object belongs to.
    def getNode(self):
        return self.node

##sets the type of the node.
##0 for internal, 1 for leaf
    def setType(self, nType):
        self.nType = nType

##Returns the type of the object
    def getType(self):
        return self.nType

##Sets the parent of the object
    def setParent(self, parent):
        self.parent = parent

##Return the parent of the object
    def getParent(self):
        return self.parent

##Sets the range of the object
    def setRange(self, range):
        self.range = range

##Returns the range of the object
    def getRange(self):
        return self.range

##Creates a copy of the object with the same parameters.
    def copy(self):
        newO = MO(self.position, self.range, self.nType)
        return newO

###########################################################


##MNode is the nodes used in the M-tree to store routing
##objects or dataobjects (MONode).
class MNode:

##Constructur. Initialises some variables.
    def __init__(self, size):
        self.size = size
        self.OrList = list()
        self.OjList = list()
        self.parent = None
        self.index = 2
        self.pd = 0

##Returns the list containing MOs. OrList has routing
##objects while OjList has dataobjects. Returns OrList
## if there are any routing objects stored, OjList if not.
    def getList(self):
        if len(self.OrList) > 0:
            return self.OrList
        else:
            return self.OjList

##Simply create empty lists for OrList and OjList.
    def emptyList(self):
        self.OrList = list()
        self.OjList = list()

##Adds a MO to either OjList or OrList based on type.
##Also sets the onde as parent to the item
    def add(self, x):
        if x.getType() == 0:
            if len(self.OrList) < self.size:
                self.OrList.append(x)
                x.setParent(self)
                return 1
        if x.getType() == 1:
            if len(self.OjList) < self.size:
                self.OjList.append(x)
                x.setParent(self)
                return 1
        return 0

##Same as above, but does not check for size.
    def add2(self, x):
        if x.getType() == 0:
            self.OrList.append(x)
            x.setParent(self)
        if x.getType() == 1:
            self.OjList.append(x)
            x.setParent(self)
        return 1

##Returns if the node is a leaf or not.
##If it has routing objects it is not.
    def isChlid(self):
        if len(self.OrList) == 0:
            return 1
        else:
            return 0

##Returns if the node is a leaf or not (redundant).
##If it has routing objects it is not.
    def isFull(self):
        if len(self.OrList) >= self.size:
            return 1
        if len(self.OjList) >= self.size:
            return 1
        return 0

##Returns if the node can store more objects or not.
    def isLeaf(self):
        if len(self.OrList) == 0:
            return 1
        else:
            return 0

##Sets an object as the parent to the node
    def setParent(self, parent):
        self.parent = parent

##Returns the parent of the node
    def getParent(self):
        return self.parent

##Replaces an object in the node with another
    def replace(self, gOp, nOp):
        for i in range(0, len(self.OrList)):
            if self.OrList[i] == gOp:
                self.OrList[i] = nOp
                self.OrList[i].setParent(self)
                break

##Removes an object from the node
    def remove(self, O):
        for i in range(0, len(self.OrList)):
            if self.OrList[i] == O:
                self.OrList.pop(i)
                break

###########################################################

##The M-tree class
class MTree:

##Constructor. Mostly Initializes a lot of global variables
    def __init__(self, pm, size, dt):
        self.range = 0
        self.size = size
        self.root = None
        self.index = 0
        self.partitionMethod = pm
        self.distanceComputations = 0
        self.condenseType = 1
        self.dt = dt
        self.dist = distance(dt)
        self.diskIO = 0
        self.dist.distanceComputations = 0
        self.samp = Samples()
        self.includeIfLess = 1

##A wrapper for inserting a new entry in the tree.
    def insertNewObject(self, entry):
        self.insert(entry)

##Inserts a new entry in the tree by creating a new object
##and then calling insert2 with the root of the tree.
    def insert(self, entry):
        entry2 = MO(entry, self.range, 1)
        if self.root == None:
            self.root = MNode(self.size)
            self.root.add(entry2)
        else:
            h = self.insert2(self.root, entry2, 0, 0)

##Recursivly calls itself until a leaf node is found.
##Chooses the branch with the closest pivot.
##If leaf empty, split is called.
    def insert2(self, node, entry, pd, level):
        self.diskIO += 1
        if node.isLeaf() == 0:
            nearestDistance = -1
            nearestNode = -1
            for i in node.getList():
                dist = self.dist.distance(i.getPosition(), entry.getPosition())
                if dist < nearestDistance or nearestDistance < 0:
                    nearestDistance = dist
                    nearestNode = i
            if nearestNode.getRange() < nearestDistance:
                nearestNode.setRange(nearestDistance)
            h = self.insert2(nearestNode.getNode(), entry, nearestDistance, level+1)
        else:
            h = level
            if node.isFull():
                self.split(node, entry)
            else:
                node.add(entry)
                entry.pd = pd
        return h

##Splits the old objects and new one into two new sets
##The pivots are chosen by the promote algorithm
##Replaces old ro with one of the new.
##If parent full, split is againc called one the new
##parent with the second ro as the entry.
    def split(self, node, entry):
        temp = node.getList()
        N = temp + [entry]
        Op = node.getParent()
        Nn = MNode(self.size)
        Nn.index = 1
        promotes = self.promote(N, self.partitionMethod)
        Op1 = promotes[0]
        Op1.setType(0)
        node.setParent(Op1)
        Op2 = promotes[1]
        Op2.setType(0)
        Nn.setParent(Op2)
        N1 = Op1.getNode()
        N2 = Op2.getNode()
        Op1.setNode(node)
        Op2.setNode(Nn)
        node.emptyList()
        oList = N1.getList()
        for i in oList:
            node.add(i)
        oList = N2.getList()
        for i in oList:
            Nn.add(i)
        if node == self.root:
            Nnp = MNode(self.size)
            Nnp.index = 0
            Nnp.add(Op1)
            Nnp.add(Op2)
            self.root = Nnp
        else:
            Np = Op.getParent()
            if Np != self.root:
                Opp = Np.getParent()
                Op1.pd = self.dist.distance(Opp.getPosition(), Op1.getPosition())
                if Opp.range < Op1.pd + Op1.range: Opp.range = Op1.pd + Op1.range
            Np.replace(Op, Op1)
            if Np.isFull():
                self.split(Np, Op2)
            else:
                if Np != self.root:
                    Op2.pd = self.dist.distance(Opp.getPosition(), Op2.getPosition())
                    if Opp.range < Op2.pd + Op2.range: Opp.range = Op2.pd + Op2.range
                Np.add(Op2)


##Just desides which promotion algorithm is to be
##called and calls it.
    def promote(self, N, promoType):
        if promoType == 0:
            return self.promoRandom(N)
        elif promoType == 1:
            return self.promoMST(N)
        elif promoType == 2:
            return self.promomMRAD(N)

##Just promotes a random pivot object. Not very
##good and possibly broken due to other changes in the code
    def promoRandom(self, N):
        rand1 = int(random.random()*len(N))
        Op1 = N[rand1].copy()
        rand2 = int(random.random()*len(N))
        while rand2 == rand1:
            rand2 = int(random.random()*len(N))
        Op2 = N[rand2].copy()
        liste = [Op1, Op2]
        return liste

##Used for promoting with the consense tree
##Only picks one routing object
    def promote2(self, N, promoType):
        if promoType == 0:
            return self.promoRandom2(N)
        if promoType == 1:
            return self.promoMRAD2(N)

##Simply pics a random or and returns it.
    def promoRandom2(self, N):
        rand = random.randint(0, len(N)-1)
        temp = N[rand].copy()
        temp.pd = N[rand].pd
        for i in N:
            if i == N[rand]:
                i.pd = 0
            else:
                i.pd = self.dist.distance(i, temp)
        return temp

##Returns the ro which gives the minimum radius.
##Also assigns the objects to it and fixes
##the parent distance and such.
    def promoMRAD2(self, N):
        d = list()
        for i in range(0, len(N)):
            d.append(list())
            for j in range(0, i+1):
                if i == j:
                    d[i].append(0)
                else:
                    d[i].append(self.dist.distance(N[i].getPosition(), N[j].getPosition()))
        bestRange = 9999
        bestNumber = 0
        for i in range(0, len(N)):
            irange = 0
            for j in range(0, i):
                if i > j:
                    if d[i][j] + N[j].range > irange:
                        d[i][j] + N[j].range
                else:
                    if d[j][i] + N[j].range > irange:
                        d[j][i] + N[j].range
            if irange < bestRange:
                bestRange = irange
                bestNumber = i
        bestp = N[bestNumber].copy()
        bestp.pd = N[bestNumber].pd
        for i in range(0, len(N)):
            if bestNumber > i:
                N[i].pd = d[bestNumber][i]
            else:
                N[i].pd = d[i][bestNumber]
        return bestp

##Partitions the objects into the different ro by
##using the GH principle (that is, the closest
##ro to the objects(, and fixes range and pd.
    def partition(self, N, Op1, Op2, N1, N2):
        for element in N:
            dist1 = self.dist.distance(Op1.getPosition(), element.getPosition())
            dist2 = self.dist.distance(Op2.getPosition(), element.getPosition())
            if dist1 < dist2:
                N1.add(element)
                if element.range > Op1.range:
                    Op1.range = element.range
                if dist1 > Op1.range:
                    Op1.range = dist1
		element.pd = dist1
            else:
                N2.add(element)
                if element.range > Op2.range:
                    Op2.range = element.range
                if dist2 > Op2.range:
                    Op2.range = dist2
                element.pd = dist2


##Used for printing the content of the tree
##Not recommended for large trees :)
    def printTree(self):
        self.counter = 0
        self.printTree2(self.root)
        print self.counter

##Recursivly calls itself downwards in the
##tree printing the content of each node.
    def printTree2(self, node):
        print "Entering node"
        if node != None:
            if node.isLeaf():
                print "Leaf"
                for element in node.getList():
                    print element.getPosition()
                    self.counter += 1
            else:
                print "Not leaf"
                Op = node.getParent()
                if Op != None:
                    print Op.position
                    print Op.range
                for element in node.getList():
                    self.printTree2(element.getNode())
        print "Returning"

##Just prints the leaves of the trees.
##The leaves contains the actual data objects
    def printLeaves(self):
        self.printLeaves(self.root)

##Recursivly prints the objects in the leaves
##in the tree
    def printLeaves2(self, node):
        if node != None:
            if node.isLeaf():
                for element in node.getList():
                    print element.getPosition()
            else:
                Op = node.getParent()
                for element in node.getList():
                    self.printTree2(element.getNode())

##Range search in the tree. Takes a query object
##and a range as parameters. Checks each of the
##roots childs and calls RS2 on them if they are
##within range.
    def rangeSearch(self, Q, r):
        self.dist.distanceComputations = 0
        self.diskIO = 0
        self.maxSearchDistance = 0
        self.minSearchDistance = 0
        Q2 = MO(Q, r, 0)
        self.result = list()
        for Or in self.root.getList():
            dist = self.dist.distance(Or.getPosition(), Q2.getPosition())
            if dist <= r + Or.getRange():
                self.RS2(Or.getNode(), Q2, r, dist)
        return self.result

##Recursive rangesearch. Calls itself downwards
##in the tree on the ro objects that are within
##range. If it's a leaf, it will check if the
##leaf is within range, and if so, include it.
    def RS2(self, node, Q, r, dist):
        self.diskIO += 1
        Op = node.getParent()
        if node.isLeaf() == 0:
            for Or in node.getList():
                if abs(dist - Or.pd) <= r + Or.getRange():
                    d = self.dist.distance(Or.getPosition(), Q.getPosition())
                    if self.includeIfLess == 1 and d <= r - Or.getRange():
                        temp = self.getAllLeaves(Or.getNode())
                        for i in temp:
                            self.result.append(i.getPosition())
                    elif d <= r + Or.getRange():
                        self.RS2(Or.getNode(), Q, r, d)
        else:
            for Oj in node.getList():
                if abs(dist - Oj.pd) <= r:
                    d = self.dist.distance(Oj.getPosition(),Q.getPosition())
                    if d <= r:
                        self.result.append(Oj.getPosition())
                        if d > self.maxSearchDistance:
                            self.maxSearchDistance = d
                        if d < self.minSearchDistance or self.minSearchDistance == 0:
                            self.minSearchDistance = d

##Used for KNN-search in the tree. Works mostly the same
##way as the range search, except that a list is created
##instead of the range with k (paremeter) spaces.
    def knnSearch(self, Q, k):
        self.dist.distanceComputations = 0
        self.diskIO = 0
        Q2 = MO(Q, 0, 0)
        self.result = [None]*k
        PR = list()
        self.k = k
        self.dk = 99999
        self.rart = 0
        for Or in self.root.getList():
            PR.append([self.dist.distance(Or.getPosition(), Q2.getPosition()), Or])
        while len(PR) > 0:
            nextNode = self.chooseNode(PR)
            self.knnSearch2(nextNode[1].getNode(), Q2, nextNode[0])
            self.updateNN(PR)
        result2 = [None]*k
        for i in range(0, k):
            result2[i] = self.result[i][1].getPosition()
        return result2
                
##Works the same way as RS2, except that it checks against
##the last (the one furthest away) object in the list
##instead of the range.
    def knnSearch2(self, node, Q, pdist):
        self.diskIO += 1
        PR = list()
        if node.isLeaf() == 0:
            for Or in node.getList():
                if abs(pdist - Or.pd) <= self.dk + Or.getRange():
                    di = self.dist.distance(Or.getPosition(), Q.getPosition())
                    if di <= self.dk + Or.getRange():
                        PR.append([di, Or])
            while len(PR) > 0:
                nextNode = self.chooseNode(PR)
                self.knnSearch2(nextNode[1].getNode(), Q, nextNode[0])
                self.updateNN(PR)
        else:
            for Oj in node.getList():
                if abs(pdist - Oj.pd) <= self.dk:
                    di = self.dist.distance(Oj.getPosition(), Q.getPosition())
                    if di < self.dk:
                        self.addNN(Oj, di)
                        
##Just pops the first node from a list. Redundant.
    def chooseNode(self, PR):
        return PR.pop(0)

##Updates a Nearest Neighbor list by removing each
##item that are outside of the minimum possible range.
    def updateNN(self, PR):
        i = 0
        while i < len(PR):
            if PR[i][0] > self.dk + PR[i][1].getRange():
                PR.pop(i)
            else:
                i += 1

##Adds an item to a Nearest Neighbour list.
##Appends the new elements, then sorts the list.
    def addNN(self, Oj, dist):
        self.result.pop(self.k-1)
        self.result.append([dist, Oj])
        self.result.sort(self.samp.noneSort)
        if self.result[self.k-1] != None:
            self.dk = self.result[self.k-1][0]

##Used for deleting an object in the tree. Mostly
##just calls delete2 with the root as parameter.
    def deleteObject(self, Object):
        self.distanceComputations = 0
        self.diskIO = 0
        DO = MO(Object, 0, 0)
        self.delete2(self.root, DO)

##Recursivly traverses the tree until it finds the
##correct leaf object, or if it runs out of potential
##objects to traverse.
##If internal, it starts with the closest ro and then
##tries everyone until object is found or no ro left.
##If leaf it removes object and returns true if it is
##found, if not returns false. If objects after delete
##is fewer than max/2 it calls condense tree.
    def delete2(self, node, DO):
        if node.isLeaf() == 0:
            sortedChildren = list()
            for i in node.getList():
                td = self.dist.distance(DO.getPosition(), i.getPosition())
                tp = [td, i]
                sortedChildren.append(tp)
            sortedChildren.sort()
            for i in sortedChildren:
                if self.delete2(i[1].getNode(), DO):
                    break
        else:
            nodeList = node.getList()
            for i in range(0, len(nodeList)):
                if nodeList[i].getPosition() == DO.getPosition():
                    nodeList.pop(i)
                    if len(nodeList) < self.size / 2:
                        self.condenseTree(node)
                    return True
            return False

##Algorithm for deleting underfilled node.
##Starts on a node, and procedes up to the root
##chekcing for underfilled nodes. If some nodes
##are underfilled, they are deleted and the objects
##are reinserted into the tree
    def condenseTree(self, node):
        X = node
        R = list()
        while X != self.root:
            Op = X.getParent()
            Np = Op.getParent()
            xList = X.getList()
            if len(xList) < self.size / 2:
                Np.remove(Op)
                R += self.getAllLeaves(X)
            else:
                Opn = self.promote2(X.getList(), self.condenseType)
                Opn.setNode(X)
                X.setParent(Opn)
                Np.replace(Op, Opn)
                Opn.range = 0
                for i in X.getList():
                    if i.pd + i.range > Opn.range:
                        Opn.range = i.pd + i.range
            X = Np
        for i in R:
            self.insert(i.getPosition())

##Returns all objects in leaves under given node.
    def getAllLeaves(self, node):
        leaves = list()
        self.diskIO += 1
        if node.isLeaf():
            leaves += node.getList()
        else:
            for i in node.getList():
                leaves += self.getAllLeaves(i.getNode())
        return leaves

##Promotes and partitions objects after the MST
##principle. First creats a distance list, then
##calls primsAlgorithm to create a MST.
##Removes the longest edge (within 90% of the longst)
##which gives most balancd cut, then cuts in two.
##Then promotes the two pivots from the two new trees
##which gives the minimum radius.
##Then it assigns the objects to the pivots and fixes
##radius, distances etc.
##BTW: This method is badly written not very efficient...
    def promoMST(self, N):
        graph = list()
        d = list()
        for i in range(0, len(N)):
            d.append(list())
            for j in range(0, i):
                if i != j:
                    di = self.dist.distance(N[i].getPosition(),N[j].getPosition())
                    dt = [di, N[i], N[j]]
                    graph.append(dt)
                    d[i].append(di)
        graph2 = copy.copy(graph)
        edges = self.primsAlgorithm(N, graph2)
        edges.sort()
        counter = 0
        bestDif = len(edges)+1
        bestseta = None
        bestsetb = None
        while counter < len(edges) and edges[0][0] >= (edges[counter][0]*0.9):
            edges2 = copy.copy(edges)
            edges2.pop(counter)
            seta = list()
            setb = list()
            e = edges2[0]
            seta.append(e[1])
            seta.append(e[2])
            allFinished = False
            while not allFinished:
                allFinished = True
                for i in range(0, len(edges2)):
                    if e[2] == edges2[i][1]:
                        seta.append(edges2[i][2])
                        e = edges2[i]
                        edges2.pop(i)
                        allFinished = False
                        break
            for i in N:
                if i not in seta:
                    setb.append(i)
            dif = abs(len(seta)-len(setb))
            if dif < bestDif:
                bestDif = dif
                bestseta = seta
                bestsetb = setb
            counter += 1

        seta = bestseta
        setb = bestsetb

        N1 = MNode(self.size)
        N2 = MNode(self.size)

        bestPivot = None
        bestRange = 999999
        for i in range(0, len(seta)):
            N1.add(seta[i])
            irange = 0
            for j in range(0, len(seta)):
                if i != j:
                    for k in range(0, len(N)):
                        if seta[i] == N[k]:
                            ii = k
                            break
                    for k in range(0, len(N)):
                        if seta[j] == N[k]:
                            jj = k
                            break
                    if ii > jj:   irange += d[ii][jj] +seta[j].getRange()
                    else:   irange += d[jj][ii] + seta[j].getRange()
            if irange < bestRange:
                bestRange = irange
                bestPivot = seta[i]
        pivot1  = copy.copy(bestPivot)
        pivot1.setRange(bestRange)
        pivot1.setNode(N1)
        for k in range(0, len(N)):
            if bestPivot == N[k]:
                pv2 = k
                break
        for i in seta:
            for k in range(0, len(N)):
                if i == N[k]:
                    if pv2 == k: i.pd = 0
                    elif pv2 > k: i.pd = d[pv2][k]
                    else: i.pd = d[k][pv2]

        bestPivot = None
        bestRange = 999999
        for i in range(0, len(setb)):
            N2.add(setb[i])
            irange = 0
            for j in range(0, len(setb)):
                if i != j:
                    for k in range(0, len(N)):
                        if setb[i] == N[k]:
                            ii = k
                            break
                    for k in range(0, len(N)):
                        if setb[j] == N[k]:
                            jj = k
                            break
                    if ii > jj:   irange += d[ii][jj] +setb[j].getRange()
                    else:   irange += d[jj][ii] + setb[j].getRange()
            if irange < bestRange:
                bestRange = irange
                bestPivot = setb[i]
        pivot2 = copy.copy(bestPivot)
        pivot2.setRange(bestRange)
        pivot2.setNode(N2)
        for k in range(0, len(N)):
            if bestPivot == N[k]:
                pv2 = k
                break
        for i in setb:
            for k in range(0, len(N)):
                if i == N[k]:
                    if pv2 == k: i.pd = 0
                    elif pv2 > k: i.pd = d[pv2][k]
                    else: i.pd = d[k][pv2]
        return [pivot1, pivot2]


##Uses prims algorithm to create a MST tree.
##Uses pricomputed distances from to avoid doing
##distance computations. Should have used a heap, but...
    def primsAlgorithm(self, N, edges):
        edges.sort()
        includedEdges = list()
        includedNodes = list()
        includedNodes.append(edges[random.randint(0, len(N)-1)][1])
        while len(includedNodes) < len(N) or len(edges) > 0:
            for i in range(0, len(edges)):
                if edges[i][1] in includedNodes and edges[i][2] not in includedNodes:
                    includedEdges.append(edges[i])
                    includedNodes.append(edges[i][2])
                    edges.pop(i)
                    break
                elif edges[i][2] in includedNodes and edges[i][1] not in includedNodes:
                    includedEdges.append(edges[i])
                    includedNodes.append(edges[i][1])
                    edges.pop(i)
                    break
                elif edges[i][1] in includedNodes and edges[i][2] in includedNodes:
                    edges.pop(i)
                    break
        return includedEdges


##Promotes the two objects that minimizes the maximum
##radius of the set of objects. Also assigns objects
##etc. Uses a distnace matrix to reduce dc.
    def promomMRAD(self, N):
        bestN1 = None
        bestN2 = None
        bestR = 99999
        best = [0]*len(N)
        distances = list()
        for i in range(0, len(N)):
            distances.append(list())
            for j in range(0, i+1):
                if i == j:
                    distances[i].append(0)
                else:
                    distances[i].append(self.dist.distance(N[i].getPosition(), N[j].getPosition()))
        for i in range(0, len(N)):
            for j in range(0, i):
                if i != j:
                    ii = N[i].copy()
                    ii.range = 0
                    jj = N[j].copy()
                    jj.range = 0
                    N1 = MNode(self.size)
                    N2 = MNode(self.size)
                    ii.setNode(N1)
                    jj.setNode(N2)
                    self.partition2(N, ii, jj, N1, N2, distances, i, j)
                    if ii.getRange() > jj.getRange() and ii.getRange() < bestR:
                        bestN1 = ii
                        bestN2 = jj
                        bestR = ii.getRange()
                        for k in range(0, len(N)):
                            best[k] = N[k].pd
                    elif jj.getRange() > ii.getRange() and jj.getRange() < bestR:
                        bestN1 = ii
                        bestN2 = jj
                        bestR = jj.getRange()
                        for k in range(0, len(N)):
                            best[k] = N[k].pd
        for k in range(0, len(N)):
            N[k].pd = best[k]
        return [bestN1, bestN2]

##Partitions the data using a distance matrix
##istead of calculating for itself after GH.
    def partition2(self, N, Op1, Op2, N1, N2, distances, t1, t2):
        for n in range(0, len(N)):
            if t1 > n:  dist1 = distances[t1][n]
            else:   dist1 = distances[n][t1]
            if t2 > n:  dist2 = distances[t2][n]
            else:   dist2 = distances[n][t2]
            if dist1 < dist2:
                N1.add(N[n])
                if dist1 + N[n].range > Op1.range:
                    Op1.range = N[n].range + dist1
                N[n].pd = dist1
            else:
                N2.add(N[n])
                if dist2 + N[n].range > Op2.range:
                    Op2.range = N[n].range + dist2
                N[n].pd = dist2

##Simply calls getAllLeaves to return objects.
##Used for compatibility.
    def getAllObjects(self, node):
        return self.getAllLeaves(node)

##Debug method for chekcing if range and parent
##distances are correct.
    def checkTree(self):
        self.checkTree2(self.root)

##Recursivly calls itself and checks every node.
    def checkTree2(self, node):
        if node.isLeaf():
            return node.getList()
        else:
            o = node.getParent()
            leaves = list()
            for i in node.getList():
                leaves += self.checkTree2(i.getNode())
            if node != self.root:
                for i in leaves:
                    d = self.dist.distance(i, o)
                    if o.range < d:
                        print "pokker"
                    else:
                        print "bra"
        return leaves

##Debug method that fixes the range and pivot
##distances of the tree. Not to be used in
##practive because it wastes dc.
    def fixTree(self):
        self.fixTree2(self.root)

##Recursivly fixes the nodes in the tree.
    def fixTree2(self, node):
        if node.isLeaf():
            return node.getList()
        else:
            o = node.getParent()
            leaves = list()
            for i in node.getList():
                leaves += self.fixTree2(i.getNode())
            if node != self.root:
                for i in leaves:
                    d = self.dist.distance(i.getPosition(), o.getPosition())
                    if o.range < d:
                        o.range = d
        return leaves

##Bulkloading algorithm from article from 1998.
##It does work, but it does hardly improve anything.
##Probably due to interpretiation error, as algorithm
##in article was hard to understand.
    def bulkLoading(self, S, M, m):
        S2 = [None]*len(S)
        for i in range(0, len(S)):
            S2[i] = MO(S[i], 0, 1)
        self.root, h = self.bulkLoading2(S2, M, m, 1)

##Because I've probably misinpreted this algorithm,
##following what is done would be a bad idea...
##But it first check if the number of objects are below a
##treshold. If it is, a tree is created of them the LEAF
##OBJECTS in them (this is probably where it is wrong).
##If not, random pivots are picked. and objects assigned.
##Underfilled pivots are deleted and objects reassigned.
##If too few objects, pivots are filled again.
##Then trees are made recursivly.
##If tree height is too low of new trees, delete and split.
##A new tree is then build bottom up from the trees
##And then call recursivly again....
    def bulkLoading2(self, S, M, m, level):
        if len(S) <= M:
            mt = MTree(self.partitionMethod, self.size, self.dt)
            liste = list()
            h = 0
            for s in S:
                if s.getNode() == None:
                    liste.append(s)
                else:
                    liste += self.getAllLeaves(s.getNode())
            for i in liste:
                mt.insert(i.getPosition())
            return mt.root, level
        else:
            F = list()
            while len(F) < 2:
                F = list()
                Ss = copy.copy(S)
                for i in range(0, M):
                    temp = Ss.pop(random.randint(0, len(Ss)-1)).copy()
                    temp.setType(0)
                    node = MNode(M)
                    temp.setNode(node)
                    node.setParent(temp)
                    F.append(temp)
                for s in S:
                    self.insertClosestVP(s, F)
                counter = 0
                while counter < len(F):
                    if len(F[counter].getNode().getList()) < m:
                        Ff = F.pop(counter)
                        for i in Ff.getNode().getList():
                            self.insertClosestVP(i, F)
                    else:
                        counter += 1
            T = list()
            hmin = 99999
            for f in F:
                Tj, h = self.bulkLoading2(f.getNode().getList(), M, m, level+1)
                f.setNode(Tj)
                Tj.setParent(f)
                T.append([Tj,h])
                if h < hmin:
                    hmin = h
            t = 0
            while t < len(T):
                if len(T[t][0].getList()) < m:
                    tF = self.splitBulk(T[t][0], 2)
                    F.pop(t)
                    T.pop(t)
                    F += tF
                else:
                    t += 1
            for t in range(0, len(T)):
                if T[t][1] > hmin:
                    tF = self.splitBulk(T[t][0], 2)
                    F.pop(t)
                    F += tF
            Tsup, h = self.bulkLoading2(F, M, m, level)
            return Tsup, level
                                
##Simply inserts an object to the closest pivot
    def insertClosestVP(self, s, F):
        bestDistance = 99999
        bestF = 0
        for f in range(0, len(F)):
            d = self.dist.distance(s.getPosition(), F[f].getPosition())
            if d < bestDistance:
                bestDistance = d
                bestF = f
        F[bestF].getNode().add2(s)
        if s.getRange() > F[bestF].getRange():
            F[bestF].setRange(s.getRange())
        s.pd = bestDistance

##Split used for bulk. Does not set range and such.
    def splitBulk(self, node, p):
        F = list()
        for i in range(0, p):
            ch = copy.copy(node.getList())
            temp = ch.pop(random.randint(0, len(ch)-1)).copy()
            temp.setType(0)
            temp2 = MNode(self.size)
            temp.setNode(temp2)
            temp2.setParent(temp)
            F.append(temp)
        for n in node.getList():
            self.insertClosestVP(n, F)
        return F

##Used for returning different properties about the tree
    def checkStuff(self):
        self.emptyLeaves = 0
        self.maxHeight = 0
        self.counter = 0
        self.checkStuff2(self.root, 1)
        return self.maxHeight, self.emptyLeaves, self.counter

##Recursivly calls itself to find usefull properties.
    def checkStuff2(self, node, height):
        self.counter += 1
        tl = node.getList()
        self.emptyLeaves += self.size - len(tl)
        if node.isLeaf():
            if height > self.maxHeight:
                self.maxHeight = height
        else:
            for i in tl:
                self.checkStuff2(i.getNode(), height + 1)
            
def feeding(tree, samples):
    for i in samples:
##        print i
        tree.insert(i)
        
def strip(x):
    result = list()
    for e in x:
        result.append(e.position)
    return result

def searchManyTimes(m, times, y, dimensions, srange):
    average = 0.0
    for i in range(0, times):
        test = samp.createSample(dimensions, 0, 1)
        noe = m.rangeSearch(test, srange)
        annet = samp.findRange(y, test, srange)
        klart = strip(noe)
        if samp.sameElements(klart, annet) == 0:
            print "Didn't match"
            return -1
        average += m.distanceComputations
    average /= times
    return average

def makeManyTrees(dimensions, centers, srange, rangeFrom, rangeTo, step, atBuild, atQuerry):
    i = rangeFrom
    for noe in range(0, len(srange)):
        atQuerry.append([[0, 0]])
    while i <= rangeTo:
        print i
        querry = [0] * len(srange)
        build = 0
        for j in range(0, 10):
            cl = samp.createClusterCenters(centers, dimensions, 0, 1)
            x = samp.createClusters(cl, i)
            y = copy.deepcopyc11(x)
            m = MTree(5, 1)
            feeding(m, x)
            build += m.distanceComputations
            for k in range(0, len(srange)):
                querry[k] += searchManyTimes(m, 100, y, dimensions, math.sqrt(srange[k]*dimensions))
        build /= 10
        build2 = [i, build]
        atBuild.append(build2)
        for k in range(0, len(srange)):
            querry[k] /= 10
            querry2 = [i, querry[k]]
            atQuerry[k].append(querry2)
        print i
        i += step
    atBuild.insert(0, [0,0])
    f = open('m - dim ' + str(dimensions) + ' cl ' + str(centers) + ' k ' + str(k) + '.txt', 'w')
    for i in atBuild:
        f.write(str(i[0]) + '  ' + str(i[1]) + '\n')
    f.close()
    for k in range(0, len(srange)):
        f = open('m - dim ' + str(dimensions) + ' cl ' + str(centers) + ' k ' + str(k) + ' ' + str(srange[k]) + '.txt', 'w')
        for i in atQuerry[k]:
            f.write(str(i[0]) + '  ' + str(i[1]) + '\n')
        f.close()
        
##c1 = [[0,0,0],[50,50,50],[100,100,100]]
##c2 = [[0,0,0],[10,10,10],[20,20,20],[30,30,30],[40,40,40],[50,50,50]]
##cl = c1
##samp = Samples()
##m = None
##wrong = False
##while not wrong:
##    cl = samp.createClusterCenters(10, 3, 0, 1)
##    x = samp.createClusters(cl, 1000)
##    m = MTree(1,10,0)
##    feeding(m, x) 
####    m.bulkLoading(x, 10, 5)
####    for i in range(0, 100):
####        temp = x.pop(random.randint(0, len(x)-1))
####        m.delete(temp)
##    m.fixTree()
####    print searchManyTimes(mvp, 100)
####    test = samp.createSample(5, 0, 1)
##    for i in range(0, 100):
##        test = samp.createSample(3, 0, 1)
##        noe = m.rangeSearch(test, 0.67082039324993692)
####        noe = m.knnSearch(test, 10)
##        annet = samp.findRange(x, test, 0.67082039324993692)
####        annet = samp.findKClosestSamples(x, test, 10)
####        klart = strip(noe)
##        klart = noe
##        if samp.sameElements(klart, annet):
##            print "wheee"
##        else:
##            print "faen"
##            print test
##            print len(klart)
##            print len(annet)
##            wrong = True
##            break
##print "ferdig"
##m.checkTree()

##srange = [0.1, 0.2, 0.3, 0.5]
##atBuild = list()
##atQuerry = list()
##makeManyTrees(10, 20, srange, 1000, 10000, 1000, atBuild, atQuerry)
##atBuild2 = list()
##atQuerry2 = list()
##makeManyTrees(10, 20, srange, 1000, 15000, 1000, atBuild2, atQuerry2, k2)
