import math
import random
from pyx import *
from pyx.graph import axis

##Utilityclass used for doing certain tasks. Not important for this project.
class Samples:

    def __init__(self):
        random.seed()

    def createUniformData(self, numberOfObjects, numberOfDimensions):
        objects = [None] * numberOfObjects
        for i in range(0, numberOfObjects):
            dim = [0] * numberOfDimensions
            for j in range(0, numberOfDimensions):
                dim[j] = random.random()
            objects[i] = dim
        return objects
    
    def boxMuller(self):
        r1 = random.random()
        r2 = random.random()
        z = math.cos(2.0*math.pi*r1)*math.sqrt(-2.0*math.log(r2))
        return z

    def normalDistributedList(self, exp, std):
        liste = list()
        numberOfElements = len(exp)
        for tall in range(0, numberOfElements):
            liste.append(exp[tall] + (std*self.boxMuller()))
        return liste

    def uniformlyDistributedList(self, centers, rangeThing):
        liste = list()
        numberOfElements = len(centers)
        for teller in range(0, numberOfElements):
            tall = centers[teller] + (random.random() * (rangeThing*2) - rangeThing)
            liste.append(tall)
        return liste

    def createClusters(self, clusterList, numberOfObjects):
        numberOfClusters = len(clusterList)
        liste = list()
        for tall in range(0, numberOfObjects):
            clusterNumber = random.randint(0,numberOfClusters-1)
            ##liste.append(self.normalDistributedList(clusterList[clusterNumber], 0.05))
            liste.append(self.uniformlyDistributedList(clusterList[clusterNumber], 0.1))
        return liste

    def createClusterCenters(self, numberOfCenters, numberOfDimensions, rangeFrom, rangeTo):
        clusterCenters = list()
        fromTo = rangeTo - rangeFrom
        for i in range(0, numberOfCenters):
            clusterCenter = list()
            for j in range(0, numberOfDimensions):
                clusterCenter.append(((random.random()*8)/10) + 0.1)
            clusterCenters.append(clusterCenter)
        return clusterCenters

    def createSample(self, numberOfDimensions, rangeFrom, rangeTo):
        sample = list()
        fromTo = rangeTo - rangeFrom
        for j in range(0, numberOfDimensions):
            sample.append(random.random()*fromTo + rangeFrom)
        return sample

    def createClusterSample(self, clusterCenters, difference):
        sample = list()
        center = random.randint(0, len(clusterCenters)-1)
        for i in clusterCenters[center]:
            dif = (random.random()*(difference*2))-difference
            sample.append(i + dif)
        return sample

    def findClosestSample(self, samples, target):
        closestNode = samples[0]
        closestDistance = self.distance(samples[0], target)
        for element in samples:
            x = self.distance(element, target)
            if x < closestDistance:
                closestDistance = x
                closestNode = element
        return closestNode

    def findKClosestSamples(self, samples, target, k):
        closestNodes = [None] * k
        closestDistances = [99999] * k
        for element in samples:
            x = self.distance(element, target)
            for i in range(0, len(closestNodes)):
                if closestNodes[i] == None or x < closestDistances[i]:
                    self.insertSorted(element, x, closestNodes, closestDistances)
                    break
        return closestNodes

    def insertSorted(self, node, x, closestNodes, closestDistances):
        for i in range(0, len(closestDistances)):
            if closestDistances[i] == None or x < closestDistances[i]:
                closestDistances.insert(i, x)
                closestDistances.pop()
                closestNodes.insert(i, node)
                closestNodes.pop()
                return

    def findRange(self, samples, target, r):
        inRange = list()
        for element in samples:
            x = self.distance(element, target)
            if x <= r:
                inRange.append(element)
        return inRange

    def distance(self, o1, o2):
        distance = 0
        for i in range(len(o1)):
##            distance += abs(o1[i] - o2[i])
##        return distance
            distance += (o1[i] - o2[i])**2
        return math.sqrt(distance)

    def sameElements(self, liste1, liste2):
##        print "sameElements"
        if len(liste1) != len(liste2):
            return 0
        holds = 1
        liste1.sort()
        liste2.sort()
        for i in range(0, len(liste1)):
            temp1 = liste1[i]
            temp2 = liste2[i]
##            print temp1
##            print temp2
            for j in range(0, len(temp1)):
                if temp1[j] != temp2[j]:
                    holds = 0
                    return holds
        return holds

    def noneSort(self, tall1, tall2):
        if tall1 == None:
            return 1
        if tall2 == None:
            return -1
        if tall1 <= tall2:
            return -1
        if tall2 > tall1:
            return 1
        else:
            return 0
        
    def generateLineGraph(self, fileNames, outputName, xValue, yValue, yType):
        if len(fileNames) > 6:
            print "Too long"
            return 0
        else:
            xText = "Data objects"
            if xValue == 0:
                xText = "Data objects"
            if xValue == 1:
                xText = "Range"
            yText = "Distance computations"
            if yValue == 0:
                yText = "Distance computations"
                yAxis = 2
            if yValue == 1:
                yText = "Disk accesses"
                yAxis = 3
            if yValue == 2:
                yText = "Objects found"
                yAxis = 4
            if yType == 0:
                g = graph.graphxy(width=8, x=axis.linear(title=xText),
                                  y=axis.linear(title=yText),
                                  key=graph.key.key(pos="tl"))
            if yType == 1:
                g = graph.graphxy(width=8, x=axis.linear(title=xText),
                                  y=axis.logarithmic(title=yText),
                                  key=graph.key.key(pos="br"))                
            test = list()
            for i in range(0, len(fileNames)):
                if i == 0:
                    d1 = graph.data.file(fileNames[0][0], x=1, y=yAxis, title=fileNames[0][1]);
                    g.plot(d1, [graph.style.line([style.linestyle.solid]), graph.style.symbol(graph.style.symbol.square, size=0.1*unit.v_cm)])
                if i == 1:
                    d2 = graph.data.file(fileNames[1][0], x=1, y=yAxis, title=fileNames[1][1]);
                    g.plot(d2, [graph.style.line([style.linestyle.dashed]), graph.style.symbol(graph.style.symbol.triangle, size=0.1*unit.v_cm)])
                if i == 2:
                    d3 = graph.data.file(fileNames[2][0], x=1, y=yAxis, title=fileNames[2][1]);
                    g.plot(d3, [graph.style.line([style.linestyle.dotted]), graph.style.symbol(graph.style.symbol.circle, size=0.1*unit.v_cm)])
                if i == 3:
                    d4 = graph.data.file(fileNames[3][0], x=1, y=yAxis, title=fileNames[3][1]);
                    g.plot(d4, [graph.style.line([style.linestyle.dashdotted]), graph.style.symbol(graph.style.symbol.diamond, size=0.1*unit.v_cm)])
                if i == 4:
                    d5 = graph.data.file(fileNames[4][0], x=1, y=yAxis, title=fileNames[4][1]);
                    g.plot(d5, [graph.style.line([style.linestyle.solid]), graph.style.symbol(graph.style.symbol.diamond, size=0.1*unit.v_cm)])
                if i == 5:
                    d5 = graph.data.file(fileNames[5][0], x=1, y=yAxis, title=fileNames[5][1]);
                    g.plot(d5, [graph.style.line([style.linestyle.dashed]), graph.style.symbol(graph.style.symbol.circle, size=0.1*unit.v_cm)])
            g.writeEPSfile(outputName)
            g.writePDFfile(outputName)
            return 1

##samp = Samples()
##fileNames = list()
##fileNames.append(["mvpq - uniform - dim 10 cl 20 k 80 range 0.1.txt", "d10r0.1"])
##fileNames.append(["mvpq - uniform - dim 10 cl 20 k 80 range 0.2.txt", "d10r0.2"])
##fileNames.append(["mvpq - uniform - dim 10 cl 20 k 80 range 0.3.txt", "d10r0.3"])
##fileNames.append(["mvpq - uniform - dim 10 cl 20 k 80 range 0.5.txt", "d10r0.5"])
##samp.generateLineGraph(fileNames, "test", 0)
