from Samples import Samples
import random
import bt
import math
import copy

##Nodes used in the MVP-tree
##Also has some variables that are used in the
##GBMVP-tree
class mvpnode(bt.gbtnode):


##Constructor that intializes a few variables.
##position is the key, PATH is the used for storing
##distances to parent nodes, children stores the
##children, M1 stores the first median, M2 stores the
##two medians to the second vp. Sv1 is the first vp,
##and Sv2 is the second vp. p is the depth.
    def __init__(self, p):
        bt.gbtnode.__init__(self)
        self.distance = 0
        self.position = None
        self.PATH = [0]*p
        self.children = list()
        self.M1 = None
        self.M2 = [None]*2
        self.D1 = 0
        self.D2 = 0
        self.D1List = list()
        self.D2List = list()
        self.Sv1 = None
        self.Sv2 = None
        self.p = 0
        self.deleted = False

##Returns all the children of the node.
    def getChildren(self):
        return self.children[0] + self.children[1] + self.children[2] + self.children[3]

##Also returns the children of the node...
    def getChildNodes(self):
        return self.children

##Replaces a child of the node
    def replaceChild(self, oldNode, newNode):
        for i in range(0, len(self.children)):
            if self.children[i] == oldNode:
                self.children[i] == newNode
                return
        print "Old node not found"


##The MVP-tree class. This class is mostly for
##creating a MVP-tree. However, due to some
##lazyness in having to rewrite more or less
##every method for the GBMVP-tree, there are some
##checks here and there that has nothing to do
##with the actual MVP-tree. But these should not influence
##the performance the MVP-tree in any way.
##For this same reason, the MVP-tree inherits the bt.gbt-
##class even if it's not a GBT-tree.
class mvptree(bt.gbt):

##Constructor. Initializes some variables.
##dtype is the type of distance function
##used, and k is the number of objects that can be stored
##in a leaf node.
    def __init__(self, k, dtype):
        bt.gbt.__init__(self, dtype)
        self.root = None
        self.k = k
        self.m = 2
        self.p = 100
        self.M1 = 0
        self.M2 = [0, 0]
        self.dist.distanceComputations = 0
        self.diskIO = 0
        self.p = 100
        self.samp = Samples()
        self.maxSearchDistance = 0
        self.minSearchDistance = 0
        self.bestDistance = 0
        self.bestDiskIO = 0
        self.worstDistance = 0
        self.worstDiskIO = 0

##Just a wrapper for building the tree from the ground.
    def insert(self, values):
        path = [0]*self.p
        self.root = self.buildTree(values,path, 0)

##Used for building a MVP-tree. Takes a set of objects,
##creats tree objects and sends them to insert2.
##The path-variable is a list of precalculated paths,
##and is used if the tree to be created is a subtree
##to already existing trees, and should therefore have
##distances to the already created pivot nodes. Level says
##which level the subtree will be on in the new tree.
    def buildTree(self, values, path, level):
##        self.dist.distanceComputations = 0
##        self.diskIO = 0
        S = list()
        for element in values:
            temp = mvpnode(self.p)
            temp.position = element
            temp.PATH = path
            temp.p = level
            S.append(temp)
        return self.insert2(S, level)

##Recursivly creates a new MVP-tree.
##If S is empty, None is returned.
##If there are less than k+2 elements in S, a new leaf node
##is created, two pivot objects (or one if only one object
##in S), are found and the rest is added to D2.
##If more than k+2 objects, two vantage points are created
##and the objects are distributed around the medians. 4 new
##subtrees are then created from these four new groups.
    def insert2(self, S, level):
        if len(S) == 0:
            return None
        node = mvpnode(self.p)        
        
        if len(S) <= self.k+2:
            Sv1 = S.pop(random.randint(0, len(S)-1))
            Sv2 = None
            if len(S) > 0:
                Sv2Number = 0
                maxDistance = 0
                D1 = list()
                for i in range(0, len(S)):
                    temp = self.dist.distance(S[i].position, Sv1.position)
                    S[i].D1 = temp
                    node.D1List.append(S[i])
                    if temp > maxDistance:
                        maxDistance = temp
                        Sv2Number = i
                Sv2 = S.pop(Sv2Number)
                for j in range(0, len(S)):
                    temp = self.dist.distance(S[j].position, Sv2.position)
                    S[j].D2 = temp
                    node.D2List.append(S[j])
            node.Sv1 = Sv1
            node.Sv2 = Sv2
            node.v = 4
            node.h = 1

        if len(S) > self.k+2:
            Sv1 = S.pop(random.randint(0, len(S)-1))
            for Si in S:
                Si.distance = self.dist.distance(Si.position, Sv1.position)
                if level < self.p:
                    Si.PATH[level] = Si.distance
                    Si.p = level
            S = self.bubbleSort(S)
            node.M1 = S[len(S)/2].distance
            SS1 = list()
            SS2 = list()
            self.part(S, SS1, SS2)
##            SS1 = S[:len(S)/2]
##            SS2 = S[len(S)/2:]
            Sv2 = SS2.pop(random.randint(0, len(SS2)-1))
            for Sj in S:
                Sj.distance = self.dist.distance(Sj.position, Sv2.position)
                if level+1 < self.p:
                    Sj.PATH[level+1] = Sj.distance
                    Sj.p = level + 1
            SM = [list(), list(), list(), list()]
            SS1 = self.bubbleSort(SS1)
##            SS1.sort(self.dist.distanceSortCriteria)  Don't ask why this isn't used instead...
            node.M2[0] = SS1[len(SS1)/2].distance
            self.part(SS1, SM[0], SM[1])
##            SM[0] = SS1[:len(SS1)/2]
##            SM[1] = SS1[len(SS1)/2:]
            SS2 = self.bubbleSort(SS2)
##            SS2.sort(self.dist.distanceSortCriteria)  Don't ask why this isn't used instead...
            node.M2[1] = SS2[len(SS2)/2].distance
            self.part(SS2, SM[2], SM[3])
##            SM[2] = SS2[:len(SS2)/2]
##            SM[3] = SS2[len(SS2)/2:]
            for i in range(0,4):
                node.children.append(self.insert2(SM[i], level + 2))
            node.Sv1 = Sv1
            node.Sv2 = Sv2
            for i in node.children:
                if i == None:
                    node.v += 1
                else:
                    node.v += i.v
                    if node.h < i.h + 2:
                        node.h = i.h + 2
        return node


##A criteria for c-sort for sorting tree-objects.
    def distanceSortCriteria(self, o1, o2):
        if o1.position < o2.position:
            return -1
        if o2.position > o1.position:
            return 1
        else:
            return 0


##The trusty old (and really bad) bubble sort algorithm!
##Except that this one only sorts half of the list
##Nice for finding medians.
    def bubbleSort(self, S):
        finished = 0
        length = len(S)-1
        length2 = len(S)/2 -1
        while finished == 0 and length > length2:
            finished = 1
            for i in range(0, length):
                if S[i].distance > S[i+1].distance:
                    temp = S[i]
                    S[i] = S[i+1]
                    S[i+1] = temp
                    finished = 0
            length -= 1
        return S

##Just partitions a list in two groups
    def part(self, S, SS1, SS2):
        for i in range(0, len(S)/2):
            SS1.append(S[i])
        for i in range(len(S)/2, len(S)):
            SS2.append(S[i])

##Prints information about the tree
##Don't use on large trees!
    def printTree(self):
        self.counter = 0
        self.vtull = 0
        self.maxh = 0
        self.printTree2(self.root, 0)
        print self.counter
        print self.maxh
        print self.vtull

##Prints information about the tree
##Don't use on large trees!
    def printTree2(self, node, h):
        if node == None: return
        h += 2
        print "Sv1"
        print node.Sv1.position
        self.counter += 1
        if node.Sv2 != None:
            print "Sv2"
            print node.Sv2.position
            self.counter += 1
        if len(node.children) > 0:
            for element in node.children:
                self.printTree2(element, h)
        else:
            self.vtull += 4
            print "D"
            for element in node.D2List:
                print element.position
                self.counter += 1
                if h > self.maxh:
                    self.maxh = h
        print "Backtracking"

##Does range search in the MVP-tree.
    def rangeSearch(self, target, searchRange):
        self.dist.distanceComputations = 0
        self.diskIO = 0
        self.ov = 0
        self.result = list()
        t2 = mvpnode(self.p)
        t2.position = target
        searchPATH = list(range(0, self.p))
        self.search2(self.root, t2, searchRange, 0, searchPATH)
        if self.dist.distanceComputations < self.bestDistance or self.bestDistance == 0:
            self.bestDistance = self.dist.distanceComputations
        if self.diskIO < self.bestDiskIO or self.bestDiskIO == 0:
            self.bestDiskIO = self.diskIO
        if self.dist.distanceComputations > self.worstDistance:
            self.worstDistance = self.dist.distanceComputations
        if self.diskIO > self.worstDiskIO:
            self.worstDiskIO = self.diskIO
        return self.result


##Recursivly range searches the MVP-tree.
##First compute distances to the vantage points and includes
##them if within range. If leaf, it checks the data objects
##in D2. Test against the PATH stored in the object for
##free pivot filtering.
##If internal, finds the appropriate childs and continues.
    def search2(self, node, Q, r, l, searchPATH):
        if node == None:
            return []
        self.diskIO += 1
        d1 = self.dist.distance(Q.position, node.Sv1.position)
        self.ov += 1
        if d1 <= r and not node.Sv1.deleted:
            self.result.append(node.Sv1.position)
            if d1 > self.maxSearchDistance:
                self.maxSearchDistance = d1
            if d1 < self.minSearchDistance or self.minSearchDistance == 0:
                self.minSearchDistance = d1
        if node.Sv2 != None:
            self.ov += 1
            d2 = self.dist.distance(Q.position, node.Sv2.position)
            if d2 <= r and not node.Sv2.deleted:
                self.result.append(node.Sv2.position)
            if d2 > self.maxSearchDistance:
                self.maxSearchDistance = d2
            if d2 < self.minSearchDistance or self.minSearchDistance == 0:
                self.minSearchDistance = d2
        if len(node.children) == 0:
            DList = node.D2List
            for Si in DList:
                self.ov += 1
                if d1 - r <= Si.D1 and Si.D1 <= d1 + r:
                    if d2 == None or (d2 - r <= Si.D2 and Si.D2 <= d2 + r):
                        holds = 1
                        for i in range(0, node.p):
                            if (searchPATH[i] - r <= Si.PATH[i] and Si.PATH[i] <= searchPATH[i] + r) == 0:
                                holds = 0
                                break
                        if holds:
                            nd = self.dist.distance(Q.position, Si.position)
                            if nd <= r:
                                self.result.append(Si.position)
                                if nd > self.maxSearchDistance:
                                    self.maxSearchDistance = nd
                                if nd < self.minSearchDistance or self.minSearchDistance == 0:
                                    self.minSearchDistance = nd
        else:
            if l < node.p:
                searchPATH[l] = d1
            if l+1 < node.p:
                searchPATH[l+1] = d2
            if d1 + node.M1 <= r:
                self.result += self.getAllObjects(node.children[0])
                self.result += self.getAllObjects(node.children[1])
            elif d1 - r <= node.M1:
                if d2 + node.M2[0] <= r:
                    self.result += self.getAllObjects(node.children[0])
                elif d2 - r <= node.M2[0]:
                    self.search2(node.children[0], Q, r, l+2, searchPATH)
                if d2 + r >= node.M2[0]:
                    self.search2(node.children[1], Q, r, l+2, searchPATH)
            if d1 + r >= node.M1:
                if d2 + node.M2[1] <= r:
                    self.result += self.getAllObjects(node.children[2])
                elif d2 - r <= node.M2[1]:
                    self.search2(node.children[2], Q, r, l+2, searchPATH)
                if d2 + r >= node.M2[1]:
                    self.search2(node.children[3], Q, r, l+2, searchPATH)

##Does kNN-search in the tree. Takes a query object and
##k which says how many objects are to be included.
    def knnSearch(self, target, k):
        self.dist.distanceComputations = 0
        self.diskIO = 0
        self.result = [None] * k
        t2 = mvpnode(self.p)
        t2.position = target
        searchPATH = list(range(0, self.p))
        self.knnSearch2(self.root, t2, k, 0, searchPATH)
        result2 = [0]*k
        for i in range(0, len(self.result)):
            result2[i] = self.result[i][1].position
        return result2

##Recursivly does kNN-search. Worsk more or less the
##same way as the range search, but checks against
##the farthest away node in the list instead of a range.
    def knnSearch2(self, node, Q, k, l, searchPATH):
        if node == None:
            return
        self.diskIO += 1
        d1 = self.dist.distance(Q.position, node.Sv1.position)
        if not node.Sv1.deleted and (self.result[k-1] == None or d1 <= self.result[k-1][0]):
            self.knnInsert(node.Sv1, d1, k)
        if node.Sv2 != None:
            d2 = self.dist.distance(Q.position, node.Sv2.position)
            if not node.Sv2.deleted and (self.result[k-1] == None or d2 <= self.result[k-1][0]):
                self.knnInsert(node.Sv2, d2, k)
        if len(node.children) == 0:
            DList = node.D2List
            for Si in DList:
                if self.result[k-1] == None or d1 - self.result[k-1][0] <= Si.D1 and Si.D1 <= d1 + self.result[k-1][0]:
                    if self.result[k-1] == None or d2 == None or (d2 - self.result[k-1][0] <= Si.D2 and Si.D2 <= d2 + self.result[k-1][0]):
                        holds = 1
                        if self.result[k-1] != None:
                            for i in range(0, node.p):
                                if (searchPATH[i] - self.result[k-1][0] <= Si.PATH[i] and Si.PATH[i] <= searchPATH[i] + self.result[k-1][0]) == 0:
                                    holds = 0
                                    break
                        if holds:
                            distance = self.dist.distance(Q.position, Si.position)
                            if self.result[k-1] == None or distance <= self.result[k-1][0]:
                                self.knnInsert(Si, distance, k)
        else:
            if l < node.p:
                searchPATH[l] = d1
            if l+1 < node.p:
                searchPATH[l+1] = d2
            if self.result[k-1] == None or d1 - self.result[k-1][0] <= node.M1:
                if self.result[k-1] == None or d2 - self.result[k-1][0] <= node.M2[0]:
                    self.knnSearch2(node.children[0], Q, k, l+2, searchPATH)
                if self.result[k-1] == None or d2 + self.result[k-1][0] >= node.M2[0]:
                    self.knnSearch2(node.children[1], Q, k, l+2, searchPATH)
            if self.result[k-1] == None or d1 + self.result[k-1][0] >= node.M1:
                if self.result[k-1] == None or d2 - self.result[k-1][0] <= node.M2[1]:
                    self.knnSearch2(node.children[2], Q, k, l+2, searchPATH)
                if self.result[k-1] == None or d2 + self.result[k-1][0] >= node.M2[1]:
                    self.knnSearch2(node.children[3], Q, k, l+2, searchPATH)
        
##Inserts a node a kNN-list and sorts it
    def knnInsert(self, mvpn, distance, k):
        element = [distance, mvpn]
        self.result[k-1] = element
        self.result.sort(self.samp.noneSort)

##Returns all objects in the tree with node as the root,
##except when the node is marked deleted (only for vps).
##This is because of the GBT.
    def getAllObjects(self, node):
        if node == None:
            return []
        self.diskIO += 1
        returnList = []
        if node.Sv1 != None:
            returnList = [node.Sv1.position]
        if node.Sv2 != None:
            returnList.append(node.Sv2.position)
        if len(node.children) == 0:
            for i in node.D2List:
                returnList.append(i.position)
        else:
            for i in node.children:
                if i != None:
                    returnList += self.getAllObjects(i)
        return returnList

##Retruns the maximum height, the number of empty
##leaves and the number of objects in the tree.
    def checkStuff(self):
        self.counter = 0
        self.emptyLeaves = 0
        self.maxh = 0
        self.checkStuff2(self.root, 0)
        return self.maxh, self.emptyLeaves, self.counter

##Retruns the maximum height, the number of empty
##leaves and the number of objects in the tree.
    def checkStuff2(self, node, h):
        if node == None: return
        h += 2
        if not node.Sv1.deleted:    self.counter += 1
        if node.Sv2 != None and not node.Sv2.deleted:
            self.counter += 1
        if len(node.children) > 0:
            for element in node.children:
                self.checkStuff2(element, h)
        else:
            self.emptyLeaves += 4
            for element in node.D2List:
                self.counter += 1
                if h > self.maxh:
                    self.maxh = h
