from Samples import Samples
from mvptree import mvptree
from mvptree import mvpnode
import math

class gbmvptree(mvptree):

##Basic constructor. Just calls the supoerc
    def __init__(self, k, dt):
        mvptree.__init__(self, k, dt)
        self.deleteMethod = 1

##Used for inserting a new object in the tree.
##Just creates a new tree object with the insert
##object as the key and then calls insertNewObject2
##with the root as a parameter.
    def insertNewObject(self, do):
        insertObject = mvpnode(self.p)
        insertObject.position = do
        self.root, insertPath = self.insertNewObject2(self.root, insertObject, 0)
        self.checkRebuild(insertPath, 2)
        
##Recursivly calls itself until a leaf node is found
##where the node can be inserted.
##Leaf:
##First checks if any vantage points are empty.
##If they are, the new node is inserted as a vantage
##point. If not, inserts into list of child objects.
##If full, the tree is rebuilt using the old objects
##as well as the new one. and new root returned.
##Internal:
##Computes distance to vantage points and selects the
##appropriate child to call itself on considereing the
##medians.
    def insertNewObject2(self, node, insertObject, level):
        if node == None:
            node = mvpnode(self.p)
            node.Sv1 = insertObject
            node.v = 4
            insertPath = [node]
            return node, insertPath
        if node.Sv2 == None:
            node.Sv2 = insertObject
            insertPath = [node]
            return node, insertPath
        else:
            d1 = self.dist.distance(node.Sv1.position, insertObject.position)
            d2 = self.dist.distance(node.Sv2.position, insertObject.position)
            if len(node.children) > 0: ##Not leaf
                if level < self.p:
                    node.PATH[level] = d1
                    node.p = level
                if  level+1 < self.p:
                    node.PATH[level+1] = d2
                    node.p = level+1
                insertPath = [node]
                test = []
                if d1 < node.M1:
                    if d2 < node.M2[0]:
                        node.children[0], test = self.insertNewObject2(node.children[0], insertObject, level+2)
                    else:
                        node.children[1], test = self.insertNewObject2(node.children[1], insertObject, level+2)
                else:
                    if d2 < node.M2[1]:
                        node.children[2], test = self.insertNewObject2(node.children[2], insertObject, level+2)
                    else:
                        node.children[3], test = self.insertNewObject2(node.children[3], insertObject, level+2)
                insertPath += test
                node.v = 0
                for i in node.children:
                    if i != None and i.h+2 > node.h:
                        node.h = i.h+2
                    if i == None:
                        node.v += 1
                    else:
                        node.v += i.v
            else: ##Leaf
                if len(node.D2List) < self.k:
                    insertObject.D1 = d1
                    insertObject.D2 = d2
                    node.D2List.append(insertObject)
                    node.v = 4
                    node.h = 1
                    insertPath = [node]
                else:
                    objects = self.getAllObjects(node)
                    objects.append(insertObject.position)
                    node = self.buildTree(objects, node.PATH, self.p)
                    insertPath = [node] 
                    for i in node.children:
                        if i != None:
                            if i.h+2 > node.h:
                                node.h = i.h+2
                            if i == None:
                                node.v += 1
                            else:
                                node.v += i.v
                        else:
                            node.v += 1
                    for i in node.children:
                        if i != None:
                            insertPath += [i]
                            break
                    
        return node, insertPath
    
##Used for deleting objects from the tree.
##Calls the appropriate delete method
    def deleteObject(self, do):
        deleteObject = mvpnode(0)
        deleteObject.position = do
##        self.distanceComputations = 0
##        self.diskIO = 0
        if self.deleteMethod == 0:
            self.root = self.deleteObject2(self.root, deleteObject)
        elif self.deleteMethod == 1:
            self.root = self.deleteObject3(self.root, deleteObject)
        self.checkRebuild([],2)


##Deletes an object using rebuilding.
##Leaf:
##If object is not found, the node is returned.
##If object is vantage point, all objects are
##gathered, except delete object. A new tee is built
##and new root returned.
##If object is in list of leaf objects it is removed
##and node is deleted.
##Internal: If object is vp, then rebuilding just like
##leaf. If not, calls recursivly on appropriate child.
    def deleteObject2(self, node, deleteObject):
        if node == None:
            print "MVP: Node not found 2"
            return None
        self.diskIO += 1
        if node.Sv1.position == deleteObject.position:
            self.updateDelete()
            return self.deleteSv(node, 0)
        elif node.Sv2.position == deleteObject.position:
            self.updateDelete()
            return self.deleteSv(node, 1)
        if len(node.children) > 0: ##Not leaf
            d1 = self.dist.distance(node.Sv1.position, deleteObject.position)
            d2 = self.dist.distance(node.Sv2.position, deleteObject.position)
            if d1 <= node.M1:
                if d2 <= node.M2[0]:
                    node.children[0] = self.deleteObject2(node.children[0], deleteObject)
                if d2 >= node.M2[0]:
                    node.children[1] = self.deleteObject2(node.children[1], deleteObject)
            if d1 >= node.M1:
                if d2 <= node.M2[1]:
                    node.children[2] = self.deleteObject2(node.children[2], deleteObject)
                if d2 >= node.M2[1]:
                    node.children[3] = self.deleteObject2(node.children[3], deleteObject)
            self.updateHV(node)
        else: ##Leaf
            for i in range(0, len(node.D2List)):
                if node.D2List[i].position == deleteObject.position:
                    node.D2List.pop(i)
                    self.updateDelete()
                    break
        return node

##Deletes an object using marking.
##Leaf:
##If object is not found, the node is returned.
##If object is vantage point, it is marked as deleted
##and the node is returned.
##If object is in list of leaf objects it is removed
##and node is deleted.
##Internal: If object is vp, then rebuilding just like
##leaf. If not, calls recursivly on appropriate child.
    def deleteObject3(self, node, deleteObject):
        if node == None:
            print "MVP: Node not found 3"
            return None
        self.diskIO += 1
        if node.Sv1.position == deleteObject.position:
            self.updateDelete()
            node.Sv1.deleted = True
            return node
        elif node.Sv2 != None and node.Sv2.position == deleteObject.position:
            self.updateDelete()
            node.Sv2.deleted = True
            return node
        if len(node.children) > 0: ##Not leaf
            d1 = self.dist.distance(node.Sv1.position, deleteObject.position)
            d2 = self.dist.distance(node.Sv2.position, deleteObject.position)
            if d1 <= node.M1:
                if d2 <= node.M2[0]:
                    node.children[0] = self.deleteObject3(node.children[0], deleteObject)
                if d2 >= node.M2[0]:
                    node.children[1] = self.deleteObject3(node.children[1], deleteObject)
            if d1 >= node.M1:
                if d2 <= node.M2[1]:
                    node.children[2] = self.deleteObject3(node.children[2], deleteObject)
                if d2 >= node.M2[1]:
                    node.children[3] = self.deleteObject3(node.children[3], deleteObject)
            self.updateHV(node)
        else: ##Leaf
            for i in range(0, len(node.D2List)):
                if node.D2List[i].position == deleteObject.position:
                    node.D2List.pop(i)
                    self.updateDelete()
                    return node
        return node
        
##Used for deleting vantage points. Gathers the nodes except
##for the vantage point given by number and rebuilds.
    def deleteSv(self, node, number):
        objects = []
        if number == 0:
            if node.Sv2 != None:
                objects = [node.Sv2.position]
        else:
            objects = [node.Sv1.position]
        for i in node.D2List:
            objects.append(i.position)
        for i in node.children:
            if i != None:   objects += self.getAllObjects(i)
        return self.buildTree(objects, node.PATH, node.p)

##Returns all objects in the tree with node as the root,
##except if the Sv's if they have been marked as deleted.
    def getAllObjects(self, node):
        if node == None:
            return []
        self.diskIO += 1
        objects = list()
        if node.Sv1 != None and not node.Sv1.deleted:
            objects = [node.Sv1.position]
        if node.Sv2 != None and not node.Sv2.deleted:
            objects.append(node.Sv2.position)
        for i in node.D2List:
            objects.append(i.position)
        for i in node.children:
            if i != None:
                objects += self.getAllObjects(i)
        return objects

##Updates h (height) and v (the number of empty leaves
##below in the nodes. v is used for rebuilding.
    def updateHV(self, node):
        for i in node.children:
            if i != None and i.h+2 > node.h:
                node.h = i.h+2
            if i == None:
                node.v += 1
            else:
                node.v += i.v

##Rebuilds the tree
    def rebuildTree(self, node):
        print "rebuild tree"
        liste = self.getAllObjects(node)
        return self.buildTree(liste, node.PATH, node.p)
