from Samples import Samples
import random
import bt
import copy
import math
from distance import distance

##The nodes used in the VP-tree
##Basically only holds a few variables that
##are use. p holds the key of the node, left
##and right holds the children and mu the median.
class VPNode(bt.gbtnode):

##Constructor. Just initializes some variables.
    def __init__(self):
        bt.gbtnode.__init__(self)
        self.p = None
        self.mu = 0
        self.left = None
        self.right = None
        self.id = None
        self.hist = None
        self.k = 0
        self.deleted = False

##Returns the two children of the node
    def getChildNodes(self):
        return [self.left, self.right]

##Replaces a childe with another object.
    def replaceChild(self, o1, o2):
        if self.left == o1:
            self.left = o2
        elif self.right == o1:
            self.right = o2
        else:
            print "Did not find object"


##The class for the VP-tree.
##Instead of subclassing the VP-tree, the code for
##the GBVP-tree was directly included in VP-tree
##because of experimentation.
##This class can act as both a VPTree and a GBVPTree.
##The bulk loading will work just as with an normal
##VP-tree, and it isn't before the insert and delete
##operations are used that the GBVP-functions are
##utilzed. However, there are some checks here and
##there that would not have been needed had this
##just been a VP-tree, but they will not inflict in
##any way.
##To get the GBT-framewor, this class inherits bt.gbt.
class VPTree(bt.gbt):

##Constructor. Just initializes some variables.
    def __init__(self, per, dtype):
        bt.gbt.__init__(self, dtype)
        random.seed(None)
        self.diskIO = 0
        self.root = None
        self.per = per
        self.maxSearchDistance = 0
        self.minSearchDistance = 0
        self.bestDistance = 0
        self.bestDiskIO = 0
        self.worstDistance = 0
        self.worstDiskIO = 0
        self.deleteMethod = 0

##Wrapper for the buildTree algorithm for compatibility.
    def insert(self, S):
##        self.dist.distanceComputations = 0
##        self.diskIO = 0
        self.root = self.buildTree(S)

##Builds a VP-tree out of list S of data objects.
##Calls itself recursivly until no objects are left
##in S.
##Uses selectVP to find a Vantage Point, removes it
##from the list and then determines the median.
##Partitions the data object into two objects around
##the median and calls itself on both storing the roots.
    def buildTree(self, S):
        if len(S) == 0:
            return None
        node = VPNode()
        if self.per > 0:
            node.p = self.selectVP(S)
        else:
            node.p = S[random.randint(0, len(S)-1)]
        self.removeElement(node.p, S)
        if len(S) > 0:
            sortedList = self.medianDistance(node.p, S)
            node.mu = sortedList[len(sortedList)/2][0]
##            print node.mu
            L = list()
            R = list()
            for i in range(0, len(sortedList)/2):
                temp = sortedList[i]
                L.append(temp[1])
            for i in range(len(sortedList)/2, len(sortedList)):
                temp = sortedList[i]
                R.append(temp[1])
            node.left = self.buildTree(L)
            node.right = self.buildTree(R)
        if node.left != None:
            node.h = node.left.h
            node.v += node.left.v
        else:
            node.v += 1
        if node.right != None:
            if node.right.h > node.h:
                node.h = node.right.h
            node.v += node.right.v
        else:
            node.v += 1
        node.h += 1
        return node

##Selects the VP depending on which potentials
##from sampleRandom that gives the largest
##variance on another sample of objects.
    def selectVP(self, S):
        ##print "Ny selectVP"
        P = self.sampleRandom(S)
        ##print len(P)
        bestSpread = 0
        bestP = P[0]
        for element in P:
            ##print "Nytt element"
            ##print element
            D = self.sampleRandom(S)
            sortedList = self.medianDistance(element, D)
            mu = sortedList[len(sortedList)/2][0]
            spread = self.spread(sortedList, mu)
            ##print spread
            if spread > bestSpread:
                bestSpread = spread
                bestP = element
        return bestP

##Retruns a list of the median distances to all the
##objects in S to p along with the objects.
    def medianDistance(self, p, S):
        ##print "Median mot:"
        if len(S) == 0:
            return 0
        distances = list()
        for s in S:
            ##print s
            distances.append([self.dist.distance(p, s), s])
        distances.sort()
        return distances

##Computes the median of a list.
    def median(self, S):
        if len(S) == 0:
            return 0
        S.sort()
        return S[len(S) / 2];
        
##Picks a random sample of objects from a list determined
##by self.per.
    def sampleRandom(self, S):
##        print "sampler"
        samples = list()
        for element in S:
            if random.random() < self.per:
                samples.append(element)
        if len(samples) == 0:
            samples.append(S[random.randint(0, len(S)-1)])
        return samples

##Computes the distance between two objects. Now obsolete.
    def distance(self, o1, o2):
        self.dist.distanceComputations += 1
        distance = 0
        for i in range(len(o1)):
            distance += (o1[i] - o2[i])**2
        return math.sqrt(distance)

##Finds the closest neighbour to a query object.
    def iNearestSearch(self, target, searchRange):
        self.dist.distanceComputations = 0
        self.tau = searchRange
        self.best = None
        self.distances = 0
        self.search2(self.root, target)
        return self.best

##Recursivly traverses a tree to find the closest
##sample. The range is always set to be the
##distance to the closest sample to improve pruning.
##Pruning is done by checking if the range of the
##search added to the distance between vp and qo
##is compared to the median.
    def iNearestSearch2(self, node, target):
        if node != None:
            x = self.dist.distance(node.p, target)
            self.distances = self.distances + 1
            ##print node.mu
            ##print x
            if x < self.tau:
                self.tau = x
                self.best = node.p
            if x - self.tau <= node.mu:
                self.search2(node.left, target)
            if x + self.tau >= node.mu:
                self.search2(node.right, target)

##Used k-Nearest Neighbour search in the tree.
##Takes a query object and the number of objects
##k to be returned, and then calls knnSearch2 to
##find them with the root as starting point.
    def knnSearch(self, target, k):
        self.dist.distanceComputations = 0
        self.diskIO = 0
        self.best = [None]*k
        self.distances = 0
        self.knnSearch2(self.root, target, k)
        best2 = [0]*k
        for i in range(0,k):
            best2[i] = self.best[i].p
        return best2

##Recursivly performs KNN-search on the tree.
##Works almost exactly like the 1NN-search,
##but now the range is the value that is the
##farthest away in the list of k objects.
    def knnSearch2(self, node, target, k):
        if node != None:
            self.diskIO +=1
            x = self.dist.distance(node.p, target)
            self.distances += 1
            node.k = x
            if not node.deleted and (self.best[k-1] == None or x < self.best[k-1].k):
                self.insertSorted(node, self.best)
            if self.best[k-1] == None or x - self.best[k-1].k <= node.mu:
                self.knnSearch2(node.left, target, k)
            if self.best[k-1] == None or x + self.best[k-1].k >= node.mu:
                self.knnSearch2(node.right, target, k)        

##Used for range search in the tree. Takes a query
##object and a range, and retrieves every object in
##the tree that is within range.
    def rangeSearch(self, target, r):
        self.dist.distanceComputations = 0
        self.diskIO = 0
        self.result = list()
        self.rangeSearch2(self.root, target, r)
        if self.dist.distanceComputations < self.bestDistance or self.bestDistance == 0:
            self.bestDistance = self.dist.distanceComputations
        if self.diskIO < self.bestDiskIO or self.bestDiskIO == 0:
            self.bestDiskIO = self.diskIO
        if self.dist.distanceComputations > self.worstDistance:
            self.worsteDistance = self.dist.distanceComputations
        if self.diskIO > self.worstDiskIO:
            self.worsteDiskIO = self.diskIO
        return self.result

##Performs recursive range search in the tree. Workst much
##like 1NN and kNN, but now every object within the range
##given at search time is included.
    def rangeSearch2(self, node, target, r):
        if node == None:
            return
        self.diskIO += 1
        x = self.dist.distance(node.p, target)
        if x <= r and not node.deleted:
            self.result.append(node.p)
            if x > self.maxSearchDistance:
                self.maxSearchDistance = x
            if x < self.minSearchDistance or self.minSearchDistance == 0:
                self.minSearchDistance = x
        if x + r >= node.mu:
            self.rangeSearch2(node.right, target, r)
        if node.left != None and x + node.mu <= r:
            self.result += self.getAllObjects(node.left)
        elif x - r <= node.mu:
            self.rangeSearch2(node.left, target, r)

##Uses insertionsort of an object into a list.
    def insertSorted(self, node, liste):
        for i in range(0, len(liste)):
            if liste[i] == None:
                liste[i] = node
                return
            if node.k < liste[i].k:
                liste.insert(i, node)
                liste.pop(len(liste)-1)
                return
                    
##Computes the spread of an object compared to a list
##of objects.
    def spread(self, D, mu):
        if len(D) < 2:
            return 0
        n = 0
        sum = 0
        sum_sqr = 0
        for element in D:
            n = n+1
            x = element[0] - mu
            sum += x
            sum_sqr += x*x
        mean = sum / n
        variance = (sum_sqr - sum*mean) / (n-1)
        return variance

##Removes a given element from a list.
    def removeElement(self, element, S):
        for number in range(len(S)):
            if S[number] == element:
                S.pop(number)
                break

##Used for printing the content of the tree.
##Not recommended for large trees :)
    def printTree2(self, node):
        if node == None:
            node = self.root
        self.counter = self.counter + 1
        print node.p
        print node.h
        if node.left != None:
            self.printTree2(node.left)
        else:
            self.empty += 1
        if node.right != None:
            self.printTree2(node.right)
        else:
            self.empty += 1
        print "backing up"


##Used for printing the content of the tree.
##Not recommended for large trees :)
    def printTree(self):
        self.counter = 0
        self.empty = 0
        self.printTree2(self.root)
        print self.counter
        print self.empty

##Returns all dataobjects in a given tree.
    def getAllObjects(self, node):
        if not node.deleted:
            liste = [node.p]
        else:
            liste = []
        self.diskIO += 1
        if node.left != None:
            liste += self.getAllObjects(node.left)
        if node.right != None:
            liste += self.getAllObjects(node.right)
        return liste

##Inserts an object into the tre. Creates a node
##of the insert object and calls findInsertNode
##with the root as a paremeter.
    def insertNewObject(self, do):
        newNode = VPNode()
        newNode.p = do
        self.root, path = self.findInsertNode(self.root, newNode)
        self.checkRebuild(path, 1)


##Recursivly calls itself until it finds an empty
##node. Insertes the new node there.
##Goes left is distance to median is less, right if
##more. When free node is found, it sets the new
##median to be the distance to the parent.
    def findInsertNode(self, node, newNode):
        path = [node]
        d = self.dist.distance(node.p, newNode.p)
        if d < node.mu:
            if node.left == None:
                node.left = newNode
                node.left.h = 1
                node.left.v = 2
                node.v += 1
                temp = [newNode]
                newNode.mu = d
            else:
                node.left, temp = self.findInsertNode(node.left, newNode)
                self.updateHV(node)
        else:
            if node.right == None:
                node.right = newNode
                node.right.h = 1
                node.right.v = 2
                node.v += 1
                temp = [newNode]
                newNode.mu = d
            else:
                node.right, temp = self.findInsertNode(node.right, newNode)
                self.updateHV(node)
        path += temp
        return node, path


##Updates the h (height) and v (free nodes benealt)
##values of the noes. h not actually used. v used
##rebuilding.
    def updateHV(self, node):
        node.h = 0
        node.v = 0
        if node.left != None:
            node.h = node.left.h
            node.v += node.left.v
        else:
            node.v += 1
        if node.right != None:
            if node.right.h > node.h:
                node.h = node.right.h
            node.v += node.right.v
        else:
            node.v += 1
        node.h += 1
        return node


##Used for deleting an object from the tree. Calls
##the appropriate delete method depending on global
##variable.
    def deleteObject(self, do):
        if self.deleteMethod == 0:        
            self.root = self.deleteObject2(self.root, do)
        elif self.deleteMethod == 1:
            self.root = self.deleteObject3(self.root, do)
        self.checkRebuild([], 1)
        
##Used for deleting an object using rebuilding.
##Calls itself downwards in the tree.
##_Uses the distance to the median to traverse the
##tree.
##If no object is found, it returns an empty node.
##If leaf is found, it returns an empty node.
##If vp is found, it gathers all the objects and
##rebuilds the tree, returning the new root.
    def deleteObject2(self, node, deleteNode):
        if node == None:
            return None
        if node.p == deleteNode:
            self.updateDelete()
            if node.left == None and node.right == None:
                return None
            else:
                objects = list()
                if node.left != None:
                    objects += self.getAllObjects(node.left)
                if node.right != None:
                    objects += self.getAllObjects(node.right)
                return self.buildTree(objects)
        else:
            d = self.dist.distance(node.p, deleteNode)
            if d <= node.mu:
                node.left = self.deleteObject2(node.left, deleteNode)
                self.updateHV(node)
            if d >= node.mu:
                node.right = self.deleteObject2(node.right, deleteNode)
                self.updateHV(node)
            return node


##Used for deleting an object using marking.
##Calls itself downwards in the tree.
##_Uses the distance to the median to traverse the
##tree.
##If no object is found, it returns an empty node.
##If leaf is found, it returns an empty node.
##If vp is found, it marks is as deleted and returns it.
##rebuilds the tree, returning the new root.
    def deleteObject3(self, node, deleteNode):
        if node == None:
            return None
        if node.p == deleteNode:
            self.updateDelete()
            if node.left == None and node.right == None:
                return None
            else:
                node.deleted = True
                return node
        else:
            d = self.dist.distance(node.p, deleteNode)
            if d <= node.mu:
                node.left = self.deleteObject3(node.left, deleteNode)
                self.updateHV(node)
            if d >= node.mu:
                node.right = self.deleteObject3(node.right, deleteNode)
                self.updateHV(node)
            return node

##Used for rebuilding the tree.
    def rebuildTree(self, node):
        liste = self.getAllObjects(node)
        return self.buildTree(liste)

##Returns some properties of the tree.
    def checkStuff(self):
        self.empty = 0
        self.counter = 0
        h = self.checkStuff2(self.root)
        return h, self.empty, self.counter

##Returns some properties of the tree.
    def checkStuff2(self, node):
        oh = 1
        height = 1
        self.counter = self.counter + 1
        if node.left != None:
            height = oh + self.checkStuff2(node.left)
        else:
            self.empty += 1
        if node.right != None:
            th = self.checkStuff2(node.right)
            if oh + th > height:
                height = oh + th
        else:
            self.empty += 1
        return height
